/*
 *************
 * Copyright *
 *************
 *
 * Copyright (C) 2010 Christoph Zauner
 */

/*
 ***********
 * License *
 ***********
 *
 * This file is part of Rihamark.
 * 
 * Rihamark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Rihamark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Rihamark.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This class executes the Testplan.
 */

package rmk.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import rmk.SPInterfaces.Algorithm;
import rmk.SPInterfaces.Attack;

/**
 * Controls the flow when execting a testplan and reports the progress.
 *
 */
public class Dispatcher {

    private TestPlan testplan;
	private Test test;
	private Task task;
	public int finishedTasks = 0;
	private Filer filer;

    public Dispatcher(TestPlan testplan, Task task) {
		this.testplan = testplan;
		this.task = task;
    }

    public void runTestplan() {

		if(testplan.isDeletePreviousResults()) {
			testplan.deleteResults();
		}

		int numOfTests = testplan.getNumOfTests();
		int numOfTasks = testplan.getNumOfJobs();

		Debug.debugPrint("Dispatcher: testplan has " + numOfTests + " tests");
		Debug.debugPrint(Debug.LogLevel.INFO, this.getClass().getName(), "testplan has " + numOfTasks + " tasks");

		// begin tests
		for (int testnum = 0; testnum < numOfTests; testnum++) {
			test = testplan.getTestsToRun().get(testnum);

			Debug.debugPrint(Debug.LogLevel.INFO, "Dispatcher: processing test with index " + testnum);
			Debug.debugPrint(Debug.LogLevel.INFO, "Dispatcher: current test has " + test.getAttacksToRun().size() + " attacks");

			String pathImageAttackDir = test.getImageAttackDir();

			filer = new Filer(test);
			filer.printFileList();

			// copy files
			File src = new File(test.getImageDir());
			File dst = new File(test.getImageAttackDir());

			try {
				prepareSrcDir(src);
			} catch (IOException ex) {
				Logger.getLogger(Dispatcher.class.getName()).log(Level.SEVERE, null, ex);
			}
			try {
				prepareDstDir(dst);
			} catch (IOException ex) {
				Logger.getLogger(Dispatcher.class.getName()).log(Level.SEVERE, null, ex);
			}

			if(test.isProcessAttackChain()) {
				Debug.debugPrint("Recursively copying " + src.getAbsolutePath() + " to " + dst.getAbsolutePath());
				try {
					copyFiles(src, dst);
				} catch (IOException ex) {
					Logger.getLogger(Dispatcher.class.getName()).log(Level.SEVERE, null, ex);
				}

				// begin attacks
				for (int attacknum = 0; attacknum < test.getAttacksToRun().size(); attacknum++) {
					Debug.debugPrint(Debug.LogLevel.DEBUG1, "Dispatcher: processing attack with index " + attacknum);

					applyAttack(test.getAttacksToRun().get(attacknum));
					
					if(task.isCancelled()) return;
					// report progress
					finishedTasks++;
					task.reportProgress(finishedTasks);
				}
				// end attacks
			}

			// start specified algorithms
			for (int algorithmidx = 0; algorithmidx < test.getAlgorithmsToRun().size(); algorithmidx++) {
				Debug.debugPrint("Dispatcher: processing algorithm with index " + algorithmidx);

				Algorithm algorithm = test.getAlgorithmsToRun().get(algorithmidx);
				Run run = new Run(test, algorithm);

				test.getRuns().add(run);
				Stopwatch stopwatch = new Stopwatch().start();
				try {

					if(test.testType.equals("Intra")) {
						intraComparison(test, algorithm, run, filer);
					} else {
						if(test.isFastInterTest() && algorithm.isAccessToHashValues()) {
							fastInterComparison(test, algorithm, run, filer);
						} else {
							slowInterComparison(test, algorithm, run, filer);

						}
					}

					if(task.isCancelled()) return;
					// report progress
					finishedTasks++;
					task.reportProgress(finishedTasks);
				} catch (IOException ex) {
					Logger.getLogger(Dispatcher.class.getName()).log(Level.SEVERE, null, ex);
				}
				stopwatch.stop();
				run.setElapsedTime(stopwatch.getElapsedTime());
				stopwatch.reset();
			}

		// end tests
		}

		Printer printer = new Printer(testplan);
		printer.print();
    }

private void intraComparison(Test test, Algorithm algorithm, Run run, Filer filer) throws IOException {

	for(int i = 0; i < filer.getNumOfFiles(); i++) {
		if(task.isCancelled()) return;

		Debug.debugPrint(Debug.LogLevel.DEBUG1, "Algorithm is processing file >" + filer.getFileOriginal(i) + "<");
		double distance = algorithm.cmpImages(filer.getFileOriginal(i), filer.getFileAttacked(i));
		Debug.debugPrint(Debug.LogLevel.DEBUG1,"Distance is " + distance);

		Result result = new Result();
		result.distance = distance;
		result.pathA = filer.getFileOriginal(i);
		result.pathB = filer.getFileAttacked(i);
		run.getResults().add(result);
	}
	Debug.debugPrint(Debug.LogLevel.DEBUG1,"Algortihm processed " + run.getResults().size() + " files");
}

/**
 * Utilizes a nCr like algorithm. This method can be used if the
 * hash functions is able to return the actual hash value to the
 * Rihamark Core. Otherwise the <code>slowInterComparison</code>
 * method must be used.
 *
 * @param test
 * @param algorithm
 * @param run
 * @param filer
 * @throws IOException
 */
private void fastInterComparison(Test test, Algorithm algorithm, Run run, Filer filer) throws IOException {

	boolean superFast = false;
	if(test.getAttacksToRun().size() == 0 && test.isProcessAttackChain()) {
		superFast = true;
	}

	// create hashes
	for(int i = 0; i < filer.getNumOfFiles(); i++) {
		if(task.isCancelled()) return;

		// create hashes
		Hash originalHash = algorithm.getHash(filer.getFileOriginal(i));
		Hash attackedHash;
		if(superFast) {
			// images in original and attacked folder are the same
			attackedHash = originalHash;
			Debug.debugPrint(Debug.LogLevel.DEBUG1, "Algorithm is hashing file >" + filer.getFileOriginal(i) + "< (super fast mode)");
		} else {
			Debug.debugPrint(Debug.LogLevel.DEBUG1, "Algorithm is hashing file >" + filer.getFileOriginal(i) + "< and >" + filer.getFileAttacked(i) + "<");
			attackedHash = algorithm.getHash(filer.getFileAttacked(i));
		}

		// add them
		filer.addFileOriginalHash(i, originalHash);
		filer.addFileAttackedHash(i, attackedHash);
	}

	// calculate distance values
	int k = 1;
	for(int i = 0; i <= filer.getNumOfFiles() - 2; i++, k++) {
		/*
		 * filer.numOfFiles() actually should be the size
		 * of the second "field". But because the two "fields" are
		 * supposed to have the same size we use the size of the
		 * first "field" here.
		 */
		for(int j = k; j <= filer.getNumOfFiles() - 1; j++) {
			Debug.debugPrint(Debug.LogLevel.DEBUG1, "Algorithm is comparing file >" + filer.getFileOriginal(i) + "< and >" + filer.getFileAttacked(j) + "<");
			double distance = algorithm.cmpHashes(filer.getFileOriginalHash(i), filer.getFileAttackedHash(j));
			Debug.debugPrint(Debug.LogLevel.DEBUG1,"Distance is " + distance);

			Result result = new Result();
			result.distance = distance;
			result.pathA = filer.getFileOriginal(i);
			result.pathB = filer.getFileAttacked(j);
			run.getResults().add(result);
		}
	}
	Debug.debugPrint(Debug.LogLevel.DEBUG1,"Algortihm processed " + run.getResults().size() + " files");
}

/**
 * Utilizes a nCr like algorithm. This method is used if the hash function
 * is only able to return a distance value for two given files to the
 * Rihamark Core.
 *
 * @param test
 * @param algorithm
 * @param run
 * @param filer
 * @throws IOException
 */
private void slowInterComparison(Test test, Algorithm algorithm, Run run, Filer filer) throws IOException {

	int k = 1;
	for(int i = 0; i <= filer.getNumOfFiles() - 2; i++, k++) {
		/*
		 * filer.numOfFiles() actually should be the size
		 * of the second "field". But because the two "fields" are
		 * supposed to have the same size we use the size of the
		 * first "field" here.
		 */
		for(int j = k; j <= filer.getNumOfFiles() - 1; j++) {
			if(task.isCancelled()) return;

			Debug.debugPrint(Debug.LogLevel.DEBUG1, "Algorithm is processing file >" + filer.getFileOriginal(i) + "< and >" + filer.getFileAttacked(j) + "<");
			double distance = algorithm.cmpImages(filer.getFileOriginal(i), filer.getFileAttacked(j));
			Debug.debugPrint(Debug.LogLevel.DEBUG1,"Distance is " + distance);

			Result result = new Result();
			result.distance = distance;
			result.pathA = filer.getFileOriginal(i);
			result.pathB = filer.getFileAttacked(j);
			run.getResults().add(result);
		}
	}
	Debug.debugPrint(Debug.LogLevel.DEBUG1,"Algortihm processed " + run.getResults().size() + " files");
}

/**
 * This function will copy files or directories from one location to another.
 * note that the source and the destination must be mutually exclusive. This
 * function can not be used to copy a directory to a sub directory of itself.
 * The function will also have problems if the destination files already exist.
 *
 * @param src -- A File object that represents the source for the copy
 * @param dest -- A File object that represnts the destination for the copy.
 * @throws IOException if unable to copy.
 */
private static void copyFiles(File src, File dest) throws IOException {
	//Check to ensure that the source is valid...
	if (!src.exists()) {
		throw new IOException("copyFiles: Can not find source: " + src.getAbsolutePath()+".");
	} else if (!src.canRead()) { //check to ensure we have rights to the source...
		throw new IOException("copyFiles: No right to source: " + src.getAbsolutePath()+".");
	}
	//is this a directory copy?
	if (src.isDirectory()) 	{
		if (!dest.exists()) { //does the destination already exist?
			//if not we need to make it exist if possible (note this is mkdirs not mkdir)
			if (!dest.mkdirs()) {
				throw new IOException("copyFiles: Could not create direcotry: " + dest.getAbsolutePath() + ".");
			}
		}
		//get a listing of files...
		String list[] = src.list();
		//copy all the files in the list.
		for (int i = 0; i < list.length; i++)
		{
			File dest1 = new File(dest, list[i]);
			File src1 = new File(src, list[i]);
			copyFiles(src1 , dest1);
		}
	} else {
		//This was not a directory, so lets just copy the file
		FileInputStream fin = null;
		FileOutputStream fout = null;
		byte[] buffer = new byte[4096]; //Buffer 4K at a time (you can change this).
		int bytesRead;
		try {
			//open the files for input and output
			fin =  new FileInputStream(src);
			fout = new FileOutputStream (dest);
			//while bytesRead indicates a successful read, lets write...
			while ((bytesRead = fin.read(buffer)) >= 0) {
				fout.write(buffer,0,bytesRead);
			}
		} catch (IOException e) { //Error copying file...
			IOException wrapper = new IOException("copyFiles: Unable to copy file: " +
						src.getAbsolutePath() + "to" + dest.getAbsolutePath()+".");
			wrapper.initCause(e);
			wrapper.setStackTrace(e.getStackTrace());
			throw wrapper;
		} finally { //Ensure that the files are closed (if they were open).
			if (fin != null) { fin.close(); }
			if (fout != null) { fout.close(); }
		}
	}
}

	private void prepareSrcDir(File dir) throws IOException {
		if(!dir.exists()) {
			if(!dir.mkdirs()) {
				throw new IOException("Could not create source directory " + dir.getAbsolutePath());
			}
		}
	}

	private void prepareDstDir(File dir) throws IOException {
		if(!dir.exists()) {
			if(!dir.mkdirs()) {
				throw new IOException("Could not create destination directory " + dir.getAbsolutePath());
			}
		}
	}

	/**
	 * Applies the given <code>Attack</code> to all the files
	 * in the <code>Filer</code>'s ""Attack Image Directory".
	 *
	 * @param attack the <code>Attack</code> to apply
	 */
	private void applyAttack(Attack attack) {
		for(int i = 0; i < filer.getNumOfFiles(); i++) {
			attack.runAttack(filer.getFileAttacked(i).toString());
			if(task.isCancelled()) return;
		}
	}

} // end of class
