/*
 *************
 * Copyright *
 *************
 *
 * Copyright (C) 2010 Christoph Zauner
 */

/*
 ***********
 * License *
 ***********
 *
 * This file is part of Rihamark.
 * 
 * Rihamark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Rihamark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Rihamark.  If not, see <http://www.gnu.org/licenses/>.
 */

package rmk.plugins.analyzers.jFree;

import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.category.LineAndShapeRenderer;
import org.jfree.chart.title.TextTitle;
import org.jfree.data.category.DefaultCategoryDataset;
import rmk.SPInterfaces.AnalyzerSwing;
import rmk.SPMenu.MenuItem;
import rmk.SPMenu.SPMenuHandler;
import rmk.core.Debug;
import rmk.core.Result;
import rmk.core.Run;
import rmk.core.Test;
import rmk.core.TestPlan;

/**
 * TODO implement an option to use the test indices as
 * labels for the x axis.
 *
 * @author Christoph Zauner
 */
public class EffectOfAttack extends AnalyzerSwing implements SPMenuHandler {

	/* start: mandatory service provider variables */
	private static String name = "Effect of Attack";
	private static String bundleName = "Rihamark Default Analyzer Plugins";
	private static String description = "FIXME";
	/* end: mandatory service provider variables */

	/* start: service provider menu children */
	private MenuItem itemDisplay;
	private MenuItem itemPDF;
	private MenuItem itemSVG;
	/* end: service provider menu children */

	private enum Function {
		ITEM_DISPLAY,
		ITEM_PDF,
		ITEM_SVG
	}

	public static final int PREF_CHART_WIDTH = 800;
	public static final int PREF_CHART_HEIGHT = 600;

	public EffectOfAttack() {
		super(
			name,
			bundleName,
			description
		);

		// menu child
		itemDisplay = new MenuItem("Display", this, Function.ITEM_DISPLAY.ordinal());
		addSPMenuChild(itemDisplay);

		// menu child
		itemPDF = new MenuItem("Save as PDF", this, Function.ITEM_PDF.ordinal());
		addSPMenuChild(itemPDF);

		// menu child
		itemSVG = new MenuItem("Save as SVG", this, Function.ITEM_SVG.ordinal());
		addSPMenuChild(itemSVG);
	}

	// begin plugin specific code

	/**
	 * This function creates a chart to visualize the effect of an attack
	 * when its parameters are varied. A line chart based on an
	 * <code>DefaultCategoryDataset</code> is used for visualization.
	 *
	 * Required format of the test plan:
	 *
	 * The test plan should consist of multiple tests. Each test should
	 * consist of the same algorithms. The attacks should be varied.
	 *
	 * @param testplan
	 */
	protected static JFreeChart visualizeEffectOfAttack(TestPlan testplan) throws InaptResultsException {

		DefaultCategoryDataset dataset = new DefaultCategoryDataset();

		for(Test test: testplan.getTestsToRun()) {

			// Runs
			for(Run run: test.getRuns()) {

				double distSum = 0;
				for(Result result: run.getResults()) {
					distSum += result.distance;
				}
				double distAvg = distSum / run.getResults().size();
				dataset.addValue(distAvg, run.getAlgorithm().getName(), test.getTestName());

			} // end Runs

		} // end Tests

		JFreeChart chart = createEffectOfAttackChart(dataset);
		return chart;
	}

  /**
	* Helper function for visualizeEffectOfAttack().
	*
	* @param dataset the data for the chart.
	*
	* @return a chart.
	*/
	private static JFreeChart createEffectOfAttackChart(DefaultCategoryDataset dataset) {
		// create the chart...
		JFreeChart chart = ChartFactory.createLineChart(
				"Effect of Attack",        // chart title
				"Name of Test",            // x axis label
				"Average Distance",        // y axis label
				dataset,                   // data
				PlotOrientation.VERTICAL,
				true,                      // include legend
				true,                      // tooltips
				false                      // urls
		);

		// NOW DO SOME OPTIONAL CUSTOMISATION OF THE CHART...

		chart.addSubtitle(new TextTitle("(Created with Rihamark)"));

		// get a reference to the plot for further customisation...
		CategoryPlot plot = (CategoryPlot) chart.getPlot();

		// background
		plot.setBackgroundPaint(Color.white);
		plot.setRangeGridlinePaint(Color.black);
		plot.setDomainGridlinePaint(Color.black);

		// y axis
		plot.getRangeAxis().setLowerBound(0);
		plot.getRangeAxis().setUpperBound(1);

		// show data points
		LineAndShapeRenderer renderer = (LineAndShapeRenderer)plot.getRenderer();
		renderer.setBaseShapesVisible(true);
		renderer.setDrawOutlines(true);
//		renderer.setUseFillPaint(true);

		// OPTIONAL CUSTOMISATION COMPLETED.

		return chart;
	}

	// end plugin specific code

	protected void displayPlot(ChartPanel chartPanel, JFrame mainFrame, int width, int height) {
		JDialog dialog = new JDialog(mainFrame);
		dialog.setLocationRelativeTo(mainFrame);
		dialog.add(chartPanel);
		dialog.setSize(width, height);
		dialog.setVisible(true);
	}

	public void executeHandler(int functionNumber) {
		if(functionNumber == Function.ITEM_DISPLAY.ordinal()) {
			handlerDisplay();
		} else if(functionNumber == Function.ITEM_PDF.ordinal()) {
			handlerPDF();
		} else if(functionNumber == Function.ITEM_SVG.ordinal()) {
			handlerSVG();
		}
	}

	private void handlerDisplay() {
		JFreeChart jFreeChart;
		try {
			jFreeChart = visualizeEffectOfAttack(userInterface.getTestPlan());
		} catch (InaptResultsException ex) {
			JOptionPane.showMessageDialog(userInterface.getMainFrame(), ex, "Error", JOptionPane.ERROR_MESSAGE);
			return;
		}
		ChartPanel chartPanel = new ChartPanel(jFreeChart);
		displayPlot(chartPanel, userInterface.getMainFrame(), PREF_CHART_WIDTH, PREF_CHART_HEIGHT);
	}

	private void handlerPDF() {
		JFreeChart jFreeChart;
		try {
			jFreeChart = visualizeEffectOfAttack(userInterface.getTestPlan());
		} catch (InaptResultsException ex) {
			JOptionPane.showMessageDialog(userInterface.getMainFrame(), ex, "Error", JOptionPane.ERROR_MESSAGE);
			return;
		}

		JFileChooser fc = new JFileChooser();
		fc.setDialogTitle("Choose a file to save the chart to");
		fc.setApproveButtonText("Save");

		int returnVal = fc.showOpenDialog(userInterface.getMainFrame());

        if (returnVal == JFileChooser.APPROVE_OPTION) {
			File file = fc.getSelectedFile();
			try {
				PrintChartPDF.saveChartAsPDF(file, jFreeChart, PREF_CHART_WIDTH, PREF_CHART_HEIGHT);
			} catch (IOException ex) {
				Logger.getLogger(EffectOfAttack.class.getName()).log(Level.SEVERE, null, ex);
				return;
			}
		} else {
            Debug.debugPrint("User clicked the cancel button.");
        }
	}

	private void handlerSVG() {
		JFreeChart jFreeChart;
		try {
			jFreeChart = visualizeEffectOfAttack(userInterface.getTestPlan());
		} catch (InaptResultsException ex) {
			JOptionPane.showMessageDialog(userInterface.getMainFrame(), ex, "Error", JOptionPane.ERROR_MESSAGE);
			return;
		}

		JFileChooser fc = new JFileChooser();
		fc.setDialogTitle("Choose a file to save the chart to");
		fc.setApproveButtonText("Save");

		int returnVal = fc.showOpenDialog(userInterface.getMainFrame());

        if (returnVal == JFileChooser.APPROVE_OPTION) {
			File file = fc.getSelectedFile();
			PrintChartSVG.saveChartAsSVG(file, jFreeChart, PREF_CHART_WIDTH, PREF_CHART_HEIGHT);
		} else {
            Debug.debugPrint("User clicked the cancel button.");
        }
	}

} // end of class
