/*
 *************
 * Copyright *
 *************
 *
 * Copyright (C) 2010 Christoph Zauner
 */

/*
 ***********
 * License *
 ***********
 *
 * This file is part of Rihamark.
 * 
 * Rihamark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Rihamark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Rihamark.  If not, see <http://www.gnu.org/licenses/>.
 */

package rmk.plugins.analyzers.jFree;

import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.title.TextTitle;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import rmk.SPInterfaces.AnalyzerSwing;
import rmk.SPMenu.MenuItem;
import rmk.SPMenu.SPMenuHandler;
import rmk.core.Debug;
import rmk.core.Result;
import rmk.core.Run;
import rmk.core.Test;
import rmk.core.TestPlan;

/**
 *
 * @author Christoph Zauner
 */
public class ROC extends AnalyzerSwing implements SPMenuHandler {

	/* start: mandatory service provider variables */
	private static String name = "ROC Chart";
	private static String bundleName = "Rihamark Default Analyzer Plugins";
	private static String description = "FIXME";
	/* end: mandatory service provider variables */

	/* start: service provider menu children */
	private MenuItem itemDisplay;
	private MenuItem itemPDF;
	private MenuItem itemSVG;
	/* end: service provider menu children */

	private enum Function {
		ITEM_DISPLAY,
		ITEM_PDF,
		ITEM_SVG
	}

	public static final int PREF_CHART_WIDTH = 600;
	public static final int PREF_CHART_HEIGHT = 600;
	public static final String GENERIC_ERR_MSG = "Error: Analyzer can't create the requested statistics out of the supplied test plan results.";

	public ROC() {
		super(
			name,
			bundleName,
			description
		);

		// menu child
		itemDisplay = new MenuItem("Display", this, Function.ITEM_DISPLAY.ordinal());
		addSPMenuChild(itemDisplay);

		// menu child
		itemPDF = new MenuItem("Save as PDF", this, Function.ITEM_PDF.ordinal());
		addSPMenuChild(itemPDF);

		// menu child
		itemSVG = new MenuItem("Save as SVG", this, Function.ITEM_SVG.ordinal());
		addSPMenuChild(itemSVG);
	}

	/**
	 * This function takes the first two tests in the test plan and creates for
	 * each algorithm a ROC curve. To create one ROC curve exactly one intra
	 * and one inter Run is needed.
	 */
	protected static JFreeChart visualizeROC (TestPlan testplan) throws InaptResultsException {
		double thresholdStep = 0.05;
		XYSeriesCollection dataset = new XYSeriesCollection();

		try {

			Test intraTest = testplan.getTestsToRun().get(0);
			Test interTest = testplan.getTestsToRun().get(1);

			for(int idx = 0; idx < intraTest.getRuns().size(); idx++) {
				Run intraRun = intraTest.getRuns().get(idx);
				Run interRun = interTest.getRuns().get(idx);

				XYSeries rocCurve = new XYSeries("Algorithm " + idx + " : " + intraRun.getAlgorithm() + " (" + intraRun.getAlgorithm().getBundleName() + ")");
				for(double threshold = 0; threshold <= 1; threshold += thresholdStep) {
					rocCurve.add(getFAR(interRun, threshold), getFRR(intraRun, threshold));
				}
				dataset.addSeries(rocCurve);
			}

		} catch (Exception ex) {
			throw new InaptResultsException(GENERIC_ERR_MSG);
		}

		JFreeChart chart = createROCChart(dataset);
		return chart;
	}

	/**
	 * Helper function for visualizeROC().
	 * Intra (authentic) tests.
	 *
	 * @param run
	 * @param threshold
	 * @return
	 */
	private static double getFRR(Run run, double threshold) {
		double frr = 0;
		int falseRejects = 0;
		int correctAccepts = 0;

		for(Result result: run.getResults()) {
			if(result.distance > threshold) {
				falseRejects++;
			} else {
				correctAccepts++;
			}
		}
		frr = (double)falseRejects / run.getResults().size();
		Debug.debugPrint(Debug.LogLevel.DEBUG2, "Calculated a FRR of " + frr + " for threshold " + threshold);

		return frr;
	}

	/**
	 * Helper function for visualizeROC().
	 * Inter (non-authentic) tests.
	 *
	 * @param run
	 * @param threshold
	 * @return
	 */
	private static double getFAR(Run run, double threshold) {
		double far = 0;
		int falseAccepts = 0;
		int correctRejects = 0;

		for(Result result: run.getResults()) {
			if(result.distance <= threshold) {
				falseAccepts++;
			} else {
				correctRejects++;
			}
		}
		far = (double)falseAccepts / run.getResults().size();
		Debug.debugPrint(Debug.LogLevel.DEBUG2, "Calculated a FAR of " + far + " for threshold " + threshold);

		return far;
	}

	/**
	* Helper function for visualizeROC().
	*
	* @param dataset the data for the chart.
	*
	* @return a chart.
	*/
	private static JFreeChart createROCChart(XYDataset dataset) {
		// create the chart...
		JFreeChart chart = ChartFactory.createXYLineChart(
				"ROC curve",                 // chart title
				"FAR (inter tests)",         // x axis label
				"FRR (intra tests)",         // y axis label
				dataset,                     // data
				PlotOrientation.VERTICAL,
				true,                        // include legend
				true,                        // tooltips
				false                        // urls
		);

		// NOW DO SOME OPTIONAL CUSTOMISATION OF THE CHART...

		chart.addSubtitle(new TextTitle("(Created with Rihamark)"));

		// get a reference to the plot for further customisation...
		XYPlot plot = (XYPlot) chart.getPlot();

		// background
		plot.setBackgroundPaint(Color.white);
		plot.setRangeGridlinePaint(Color.black);
		plot.setDomainGridlinePaint(Color.black);

		// x axis
		plot.getDomainAxis().setLowerBound(0);
		plot.getDomainAxis().setUpperBound(1);
		// y axis
		plot.getRangeAxis().setLowerBound(0);
		plot.getRangeAxis().setUpperBound(1);

		// OPTIONAL CUSTOMISATION COMPLETED.

		return chart;
	}

	protected void displayPlot(ChartPanel chartPanel, JFrame mainFrame, int width, int height) {
		JDialog dialog = new JDialog(mainFrame);
		dialog.setLocationRelativeTo(mainFrame);
		dialog.add(chartPanel);
		dialog.setSize(width, height);
		dialog.setVisible(true);
	}

	public void executeHandler(int functionNumber) {
		if(functionNumber == Function.ITEM_DISPLAY.ordinal()) {
			handlerDisplay();
		} else if(functionNumber == Function.ITEM_PDF.ordinal()) {
			handlerPDF();
		} else if(functionNumber == Function.ITEM_SVG.ordinal()) {
			handlerSVG();
		}
	}

	private void handlerDisplay() {
		JFreeChart jFreeChart;
		try {
			jFreeChart = visualizeROC(userInterface.getTestPlan());
		} catch (InaptResultsException ex) {
			JOptionPane.showMessageDialog(userInterface.getMainFrame(), ex, "Error", JOptionPane.ERROR_MESSAGE);
			return;
		}
		ChartPanel chartPanel = new ChartPanel(jFreeChart);
		displayPlot(chartPanel, userInterface.getMainFrame(), PREF_CHART_WIDTH, PREF_CHART_HEIGHT);
	}

	private void handlerPDF() {
		JFreeChart jFreeChart;
		try {
			jFreeChart = visualizeROC(userInterface.getTestPlan());
		} catch (InaptResultsException ex) {
			JOptionPane.showMessageDialog(userInterface.getMainFrame(), ex, "Error", JOptionPane.ERROR_MESSAGE);
			return;
		}

		JFileChooser fc = new JFileChooser();
		fc.setDialogTitle("Choose a file to save the chart to");
		fc.setApproveButtonText("Save");

		int returnVal = fc.showOpenDialog(userInterface.getMainFrame());

        if (returnVal == JFileChooser.APPROVE_OPTION) {
			File file = fc.getSelectedFile();
			try {
				PrintChartPDF.saveChartAsPDF(file, jFreeChart, PREF_CHART_WIDTH, PREF_CHART_HEIGHT);
			} catch (IOException ex) {
				Logger.getLogger(ROC.class.getName()).log(Level.SEVERE, null, ex);
				return;
			}
		} else {
            Debug.debugPrint("User clicked the cancel button.");
        }
	}

	private void handlerSVG() {
		JFreeChart jFreeChart;
		try {
			jFreeChart = visualizeROC(userInterface.getTestPlan());
		} catch (InaptResultsException ex) {
			JOptionPane.showMessageDialog(userInterface.getMainFrame(), ex, "Error", JOptionPane.ERROR_MESSAGE);
			return;
		}

		JFileChooser fc = new JFileChooser();
		fc.setDialogTitle("Choose a file to save the chart to");
		fc.setApproveButtonText("Save");

		int returnVal = fc.showOpenDialog(userInterface.getMainFrame());

        if (returnVal == JFileChooser.APPROVE_OPTION) {
			File file = fc.getSelectedFile();
			PrintChartSVG.saveChartAsSVG(file, jFreeChart, PREF_CHART_WIDTH, PREF_CHART_HEIGHT);
		} else {
            Debug.debugPrint("User clicked the cancel button.");
        }
	}

} // end of class
